@@ -63,7 +63,7 @@ except NameError:
     unicode = str
 
 
-__version__ = '0.9.3'
+__version__ = '0.9.4'
 
 
 CR = '\r'
@@ -141,6 +141,143 @@ def extended_blank_lines(logical_line,
 pep8.register_check(extended_blank_lines)
 
 
+def continued_indentation(logical_line, tokens, indent_level, noqa):
+    r"""Override pep8's function to provide indentation information."""
+    first_row = tokens[0][2][0]
+    nrows = 1 + tokens[-1][2][0] - first_row
+    if noqa or nrows == 1:
+        return
+
+    # indent_next tells us whether the next block is indented; assuming
+    # that it is indented by 4 spaces, then we should not allow 4-space
+    # indents on the final continuation line; in turn, some other
+    # indents are allowed to have an extra 4 spaces.
+    indent_next = logical_line.endswith(':')
+
+    row = depth = 0
+    # remember how many brackets were opened on each line
+    parens = [0] * nrows
+    # relative indents of physical lines
+    rel_indent = [0] * nrows
+    # visual indents
+    indent_chances = {}
+    last_indent = tokens[0][2]
+    indent = [last_indent[1]]
+
+    last_token_multiline = None
+    line = None
+    for token_type, text, start, end, line in tokens:
+
+        newline = row < start[0] - first_row
+        if newline:
+            row = start[0] - first_row
+            newline = (not last_token_multiline and
+                       token_type not in (tokenize.NL, tokenize.NEWLINE))
+
+        if newline:
+            # this is the beginning of a continuation line.
+            last_indent = start
+
+            # record the initial indent.
+            rel_indent[row] = pep8.expand_indent(line) - indent_level
+
+            if depth:
+                # a bracket expression in a continuation line.
+                # find the line that it was opened on
+                for open_row in range(row - 1, -1, -1):
+                    if parens[open_row]:
+                        break
+            else:
+                # an unbracketed continuation line (ie, backslash)
+                open_row = 0
+            hang = rel_indent[row] - rel_indent[open_row]
+            close_bracket = (token_type == tokenize.OP and text in ']})')
+            visual_indent = (not close_bracket and hang > 0 and
+                             indent_chances.get(start[1]))
+
+            if close_bracket and indent[depth]:
+                # closing bracket for visual indent
+                if start[1] != indent[depth]:
+                    yield (start, 'E124 {0}'.format(indent[depth]))
+            elif close_bracket and not hang:
+                pass
+            elif visual_indent is True:
+                # visual indent is verified
+                if not indent[depth]:
+                    indent[depth] = start[1]
+            elif visual_indent in (text, str):
+                # ignore token lined up with matching one from a previous line
+                pass
+            elif indent[depth] and start[1] < indent[depth]:
+                # visual indent is broken
+                yield (start, 'E128 {0}'.format(indent[depth]))
+            elif hang == 4 or (indent_next and rel_indent[row] == 8):
+                # hanging indent is verified
+                if close_bracket:
+                    yield (start, 'E123 {0}'.format(indent_level +
+                                                    rel_indent[open_row]))
+            else:
+                one_indented = indent_level + rel_indent[open_row] + 4
+                # indent is broken
+                if hang <= 0:
+                    error = ('E122', one_indented)
+                elif indent[depth]:
+                    error = ('E127', indent[depth])
+                elif hang % 4:
+                    error = ('E121', one_indented)
+                else:
+                    error = ('E126', one_indented)
+                yield (start, '{0} {1}'.format(*error))
+
+        # look for visual indenting
+        if (parens[row] and token_type not in (tokenize.NL, tokenize.COMMENT)
+                and not indent[depth]):
+            indent[depth] = start[1]
+            indent_chances[start[1]] = True
+        # deal with implicit string concatenation
+        elif (token_type in (tokenize.STRING, tokenize.COMMENT) or
+              text in ('u', 'ur', 'b', 'br')):
+            indent_chances[start[1]] = str
+        # special case for the "if" statement because len("if (") == 4
+        elif not indent_chances and not row and not depth and text == 'if':
+            indent_chances[end[1] + 1] = True
+
+        # keep track of bracket depth
+        if token_type == tokenize.OP:
+            if text in '([{':
+                depth += 1
+                indent.append(0)
+                parens[row] += 1
+            elif text in ')]}' and depth > 0:
+                # parent indents should not be more than this one
+                prev_indent = indent.pop() or last_indent[1]
+                for d in range(depth):
+                    if indent[d] > prev_indent:
+                        indent[d] = 0
+                for ind in list(indent_chances):
+                    if ind >= prev_indent:
+                        del indent_chances[ind]
+                depth -= 1
+                if depth:
+                    indent_chances[indent[depth]] = True
+                for idx in range(row, -1, -1):
+                    if parens[idx]:
+                        parens[idx] -= 1
+                        rel_indent[row] = rel_indent[idx]
+                        break
+            assert len(indent) == depth + 1
+            if start[1] not in indent_chances:
+                # allow to line up tokens
+                indent_chances[start[1]] = text
+
+        last_token_multiline = (start[0] != end[0])
+
+    if indent_next and pep8.expand_indent(line) == indent_level + 4:
+        yield (last_indent, 'E125 {0}'.format(indent_level + 8))
+del pep8._checks['logical_line'][pep8.continued_indentation]
+pep8.register_check(continued_indentation)
+
+
 class FixPEP8(object):
 
     """Fix invalid code.
@@ -192,8 +329,15 @@ class FixPEP8(object):
 
         # method definition
         self.fix_e111 = self.fix_e101
-        self.fix_e128 = self.fix_e127
-        self.fix_e129 = self.fix_e125
+        self.fix_e121 = self._fix_reindent
+        self.fix_e122 = self._fix_reindent
+        self.fix_e123 = self._fix_reindent
+        self.fix_e124 = self._fix_reindent
+        self.fix_e125 = self._fix_reindent
+        self.fix_e126 = self._fix_reindent
+        self.fix_e127 = self._fix_reindent
+        self.fix_e128 = self._fix_reindent
+        self.fix_e129 = self._fix_reindent
         self.fix_e202 = self.fix_e201
         self.fix_e203 = self.fix_e201
         self.fix_e211 = self.fix_e201
@@ -227,10 +371,6 @@ class FixPEP8(object):
                 line_index = result['line'] - 1
                 original_line = self.source[line_index]
 
-                # TODO: Make below logical fix stuff more efficient. Can we
-                # salvage things such that we can still apply logical fixes
-                # even though there are modified lines?
-
                 is_logical_fix = len(inspect.getargspec(fix).args) > 2
                 if is_logical_fix:
                     # Do not run logical fix if any lines have been modified.
@@ -365,145 +505,24 @@ class FixPEP8(object):
         original = self.source[ls[0]:le[0] + 1]
         return ls, le, original
 
-    def _fix_reindent(self, result, logical):
+    def _fix_reindent(self, result):
         """Fix a badly indented line.
 
         This is done by adding or removing from its initial indent only.
 
         """
-        assert logical
-        ls, _, original = logical
-
-        rewrapper = Wrapper(original)
-        valid_indents = rewrapper.pep8_expected()
-        if not rewrapper.rel_indent:
-            return []  # pragma: no cover
-        if result['line'] > ls[0]:
-            # got a valid continuation line number from pep8
-            row = result['line'] - ls[0] - 1
-            # always pick the first option for this
-            valid = valid_indents[row]
-            got = rewrapper.rel_indent[row]
-        else:
-            return []  # pragma: no cover
-        line = ls[0] + row
-        # always pick the expected indent, for now.
-        indent_to = valid[0]
-
-        if got != indent_to:
-            orig_line = self.source[line]
-            new_line = ' ' * (indent_to) + orig_line.lstrip()
-            if new_line == orig_line:
-                return []
-            else:
-                self.source[line] = new_line
-                return [line + 1]  # Line indexed at 1
-        else:
-            return []  # pragma: no cover
-
-    def fix_e121(self, result, logical):
-        """Fix indentation to be a multiple of four."""
-        # Fix by adjusting initial indent level.
-        return self._fix_reindent(result, logical)
-
-    def fix_e122(self, result, logical):
-        """Add absent indentation for hanging indentation."""
-        # Fix by adding an initial indent.
-        modified_lines = self._fix_reindent(result, logical)
-        if modified_lines:
-            return modified_lines
-        else:
-            # Fallback
-            line_index = result['line'] - 1
-            original_line = self.source[line_index]
-            indentation = _get_indentation(original_line)
-            self.source[line_index] = (indentation + self.indent_word +
-                                       original_line.lstrip())
-
-    def fix_e123(self, result, logical):
-        """Align closing bracket to match opening bracket."""
-        # Fix by deleting whitespace to the correct level.
-        assert logical
-        logical_lines = logical[2]
+        num_indent = int(result['info'].split()[1])
         line_index = result['line'] - 1
-        original_line = self.source[line_index]
+        target = self.source[line_index]
 
-        fixed_line = (_get_indentation(logical_lines[0]) +
-                      original_line.lstrip())
-        if fixed_line == original_line:
-            # Fall back to slower method.
-            return self._fix_reindent(result, logical)
-        else:
-            self.source[line_index] = fixed_line
-
-    def fix_e124(self, result, logical):
-        """Align closing bracket to match visual indentation."""
-        # Fix by inserting whitespace before the closing bracket.
-        return self._fix_reindent(result, logical)
-
-    def fix_e125(self, result, logical):
-        """Indent to distinguish line from next logical line."""
-        # Fix by indenting the line in error to the next stop.
-        return self._fix_reindent(result, logical)
-
-    def fix_e126(self, result, logical):
-        """Fix over-indented hanging indentation."""
-        # fix by deleting whitespace to the left
-        assert logical
-        logical_lines = logical[2]
-        line_index = result['line'] - 1
-        original = self.source[line_index]
-
-        fixed = (_get_indentation(logical_lines[0]) +
-                 self.indent_word + original.lstrip())
-        if fixed == original:
-            # Fall back to slower method.
-            return self._fix_reindent(result, logical)  # pragma: no cover
-        else:
-            self.source[line_index] = fixed
-
-    def fix_e127(self, result, logical):
-        """Fix visual indentation."""
-        # Fix by inserting/deleting whitespace to the correct level.
-        modified_lines = self._align_visual_indent(result, logical)
-        if modified_lines != []:
-            return modified_lines
-        else:
-            # Fall back to slower method.
-            return self._fix_reindent(result, logical)
-
-    def _align_visual_indent(self, result, logical):
-        """Correct visual indent.
-
-        This includes over (E127) and under (E128) indented lines.
-
-        """
-        assert logical
-        logical_lines = logical[2]
-        line_index = result['line'] - 1
-        original = self.source[line_index]
-        fixed = original
-
-        if logical_lines[0].rstrip().endswith('\\'):
-            fixed = (_get_indentation(logical_lines[0]) +
-                     self.indent_word + original.lstrip())
-        else:
-            start_index = None
-            for symbol in '([{':
-                if symbol in logical_lines[0]:
-                    found_index = logical_lines[0].find(symbol)
-                    if start_index is None:
-                        start_index = found_index
-                    else:
-                        start_index = min(start_index, found_index)
-
-            if start_index is not None:
-                fixed = start_index * ' ' + original.lstrip()
-
-        if fixed == original:
+        # When multiline strings are involved, pep8 reports the error as
+        # being at the start of the multiline string, which doesn't work
+        # for us.
+        if ('"""' in target or "'''" in target):
             return []
-        else:
-            self.source[line_index] = fixed
+
+        self.source[line_index] = (
+            ' ' * num_indent + target.lstrip())
 
     def fix_e201(self, result):
         """Remove extraneous whitespace."""
@@ -777,6 +796,7 @@ class FixPEP8(object):
                         _get_indentation(target) + self.indent_word +
                         target[c:].lstrip('\n\r \t\\'))
         self.source[result['line'] - 1] = fixed_source
+        return [result['line'], result['line'] + 1]
 
     def fix_e702(self, result, logical):
         """Put semicolon-separated compound statement on separate lines."""
@@ -1040,8 +1060,8 @@ def get_diff_text(old, new, filename):
 def _priority_key(pep8_result):
     """Key for sorting PEP8 results.
 
-    Global fixes should be done first. This is important for things
-    like indentation.
+    Global fixes should be done first. This is important for things like
+    indentation.
 
     """
     priority = [
@@ -1258,8 +1278,8 @@ def _execute_pep8(pep8_options, source):
         def full_error_results(self):
             """Return error results in detail.
 
-            Results are in the form of a list of dictionaries. Each dictionary
-            contains 'id', 'line', 'column', and 'info'.
+            Results are in the form of a list of dictionaries. Each
+            dictionary contains 'id', 'line', 'column', and 'info'.
 
             """
             return self.__full_error_results
@@ -1453,229 +1473,6 @@ def reindent_stats(tokens):
     return stats
 
 
-class Wrapper(object):
-
-    """Class for functions relating to continuation lines and line folding.
-
-    Each instance operates on a single logical line.
-
-    """
-
-    SKIP_TOKENS = frozenset([
-        tokenize.COMMENT, tokenize.NL, tokenize.INDENT,
-        tokenize.DEDENT, tokenize.NEWLINE, tokenize.ENDMARKER
-    ])
-
-    def __init__(self, physical_lines):
-        self.lines = physical_lines
-        self.tokens = []
-        self.rel_indent = None
-        sio = io.StringIO(''.join(physical_lines))
-        for t in tokenize.generate_tokens(sio.readline):
-            if not len(self.tokens) and t[0] in self.SKIP_TOKENS:
-                continue
-            if t[0] != tokenize.ENDMARKER:
-                self.tokens.append(t)
-
-        self.logical_line = self.build_tokens_logical(self.tokens)
-
-    def build_tokens_logical(self, tokens):
-        """Build a logical line from a list of tokens.
-
-        Return the logical line and a list of (offset, token) tuples. Does
-        not mute strings like the version in pep8.py.
-
-        """
-        # from pep8.py with minor modifications
-        logical = []
-        previous = None
-        for t in tokens:
-            token_type, text = t[0:2]
-            if token_type in self.SKIP_TOKENS:
-                continue
-            if previous:
-                end_line, end = previous[3]
-                start_line, start = t[2]
-                if end_line != start_line:  # different row
-                    prev_text = self.lines[end_line - 1][end - 1]
-                    if prev_text == ',' or (prev_text not in '{[('
-                                            and text not in '}])'):
-                        logical.append(' ')
-                elif end != start:  # different column
-                    fill = self.lines[end_line - 1][end:start]
-                    logical.append(fill)
-            logical.append(text)
-            previous = t
-        logical_line = ''.join(logical)
-        assert logical_line.lstrip() == logical_line
-        assert logical_line.rstrip() == logical_line
-        return logical_line
-
-    def pep8_expected(self):
-        """Replicate logic in pep8.py, to know what level to indent things to.
-
-        Return a list of lists; each list represents valid indent levels for
-        the line in question, relative from the initial indent. However, the
-        first entry is the indent level which was expected.
-
-        """
-        # What follows is an adjusted version of
-        # pep8.py:continuation_line_indentation. All of the comments have been
-        # stripped and the 'yield' statements replaced with 'pass'.
-        if not self.tokens:
-            return  # pragma: no cover
-
-        first_row = self.tokens[0][2][0]
-        nrows = 1 + self.tokens[-1][2][0] - first_row
-
-        # here are the return values
-        valid_indents = [list()] * nrows
-        indent_level = self.tokens[0][2][1]
-        valid_indents[0].append(indent_level)
-
-        if nrows == 1:
-            # bug, really.
-            return valid_indents  # pragma: no cover
-
-        indent_next = self.logical_line.endswith(':')
-
-        row = depth = 0
-        parens = [0] * nrows
-        self.rel_indent = rel_indent = [0] * nrows
-        indent = [indent_level]
-        indent_chances = {}
-        last_indent = (0, 0)
-        last_token_multiline = None
-
-        for token_type, text, start, end, line in self.tokens:
-            newline = row < start[0] - first_row
-            if newline:
-                row = start[0] - first_row
-                newline = (not last_token_multiline and
-                           token_type not in (tokenize.NL, tokenize.NEWLINE))
-
-            if newline:
-                # This is where the differences start. Instead of looking at
-                # the line and determining whether the observed indent matches
-                # our expectations, we decide which type of indentation is in
-                # use at the given indent level, and return the offset. This
-                # algorithm is susceptible to "carried errors", but should
-                # through repeated runs eventually solve indentation for
-                # multiline expressions less than PEP8_PASSES_MAX lines long.
-
-                if depth:
-                    for open_row in range(row - 1, -1, -1):
-                        if parens[open_row]:
-                            break
-                else:
-                    open_row = 0
-
-                # That's all we get to work with. This code attempts to
-                # "reverse" the below logic, and place into the valid indents
-                # list
-                vi = []
-                add_second_chances = False
-                if token_type == tokenize.OP and text in ']})':
-                    # this line starts with a closing bracket, so it needs to
-                    # be closed at the same indent as the opening one.
-                    if indent[depth]:
-                        # hanging indent
-                        vi.append(indent[depth])
-                    else:
-                        # visual indent
-                        vi.append(indent_level + rel_indent[open_row])
-                elif depth and indent[depth]:
-                    # visual indent was previously confirmed.
-                    vi.append(indent[depth])
-                    add_second_chances = True
-                elif depth and True in indent_chances.values():
-                    # visual indent happened before, so stick to
-                    # visual indent this time.
-                    if depth > 1 and indent[depth - 1]:
-                        vi.append(indent[depth - 1])
-                    else:
-                        # stupid fallback
-                        vi.append(indent_level + 4)
-                    add_second_chances = True
-                elif not depth:
-                    vi.append(indent_level + 4)
-                else:
-                    # must be in hanging indent
-                    hang = rel_indent[open_row] + 4
-                    vi.append(indent_level + hang)
-
-                # about the best we can do without look-ahead
-                if (indent_next and vi[0] == indent_level + 4 and
-                        nrows == row + 1):
-                    vi[0] += 4
-
-                if add_second_chances:
-                    # visual indenters like to line things up.
-                    min_indent = vi[0]
-                    for col, what in indent_chances.items():
-                        if col > min_indent and (
-                            what is True or
-                            (what == str and token_type == tokenize.STRING) or
-                            (what == text and token_type == tokenize.OP)
-                        ):
-                            vi.append(col)
-                    vi = sorted(vi)
-
-                valid_indents[row] = vi
-
-                # Returning to original continuation_line_indentation() from
-                # pep8.
-                visual_indent = indent_chances.get(start[1])
-                last_indent = start
-                rel_indent[row] = pep8.expand_indent(line) - indent_level
-                hang = rel_indent[row] - rel_indent[open_row]
-
-                if token_type == tokenize.OP and text in ']})':
-                    pass
-                elif visual_indent is True:
-                    if not indent[depth]:
-                        indent[depth] = start[1]
-
-            # line altered: comments shouldn't define a visual indent
-            if parens[row] and not indent[depth] and token_type not in (
-                tokenize.NL, tokenize.COMMENT
-            ):
-                indent[depth] = start[1]
-                indent_chances[start[1]] = True
-            elif token_type == tokenize.STRING or text in (
-                'u', 'ur', 'b', 'br'
-            ):
-                indent_chances[start[1]] = str
-
-            if token_type == tokenize.OP:
-                if text in '([{':
-                    depth += 1
-                    indent.append(0)
-                    parens[row] += 1
-                elif text in ')]}' and depth > 0:
-                    prev_indent = indent.pop() or last_indent[1]
-                    for d in range(depth):
-                        if indent[d] > prev_indent:
-                            indent[d] = 0
-                    for ind in list(indent_chances):
-                        if ind >= prev_indent:
-                            del indent_chances[ind]
-                    depth -= 1
-                    if depth and indent[depth]:  # modified
-                        indent_chances[indent[depth]] = True
-                    for idx in range(row, -1, -1):
-                        if parens[idx]:
-                            parens[idx] -= 1
-                            break
-                assert len(indent) == depth + 1
-                if start[1] not in indent_chances:
-                    indent_chances[start[1]] = text
-
-            last_token_multiline = (start[0] != end[0])
-
-        return valid_indents
-
-
 def _leading_space_count(line):
     """Return number of leading spaces in line."""
     i = 0
@@ -1937,10 +1734,8 @@ def fix_string(source, options=None):
     if not options:
         options = parse_args([''])[0]
 
-    try:
+    if not isinstance(source, unicode):
         source = source.decode(locale.getpreferredencoding(False))
-    except AttributeError:
-        pass
 
     sio = io.StringIO(source)
     return fix_lines(sio.readlines(), options=options)
@@ -2036,8 +1831,8 @@ def global_fixes():
 def apply_global_fixes(source, options):
     """Run global fixes on source code.
 
-    These are fixes that only need be done once (unlike those in FixPEP8,
-    which are dependent on pep8).
+    These are fixes that only need be done once (unlike those in
+    FixPEP8, which are dependent on pep8).
 
     """
     for (code, function) in global_fixes():
@@ -2120,6 +1915,8 @@ def parse_args(args):
     if not len(args) and not options.list_fixes:
         parser.error('incorrect number of arguments')
 
+    args = [decode_filename(name) for name in args]
+
     if '-' in args:
         if len(args) > 1:
             parser.error('cannot mix stdin and regular files')
@@ -2179,10 +1976,19 @@ def parse_args(args):
     return options, args
 
 
+def decode_filename(filename):
+    """Return Unicode filename."""
+    if isinstance(filename, unicode):
+        return filename
+    else:
+        return filename.decode(sys.getfilesystemencoding())
+
+
 def supported_fixes():
     """Yield pep8 error codes that autopep8 fixes.
 
-    Each item we yield is a tuple of the code followed by its description.
+    Each item we yield is a tuple of the code followed by its
+    description.
 
     """
     instance = FixPEP8(filename=None, options=None, contents='')
@@ -2213,7 +2019,7 @@ def line_shortening_rank(candidate, newline, indent_word):
             for symbol in '([{':
                 offset = max(offset, 1 + lines[0].find(symbol))
 
-        max_length = max([offset + len(x.strip()) for x in lines])
+        max_length = max(offset + len(x.strip()) for x in lines)
         rank += max_length
         rank += len(lines)
 
